
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Calendar;
import java.util.List;

import org.apache.pdfbox.exceptions.SignatureException;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.PDSignature;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.SignatureInterface;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.SignatureOptions;
import org.bouncycastle.cert.jcajce.JcaCertStore;
import org.bouncycastle.cms.CMSException;
import org.bouncycastle.cms.CMSProcessable;
import org.bouncycastle.cms.CMSSignedData;
import org.bouncycastle.cms.CMSSignedDataGenerator;
import org.bouncycastle.cms.jcajce.JcaSignerInfoGeneratorBuilder;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;
import org.bouncycastle.operator.jcajce.JcaDigestCalculatorProviderBuilder;

/**
 * Signs a pdf document with a certificate. This class doesn't support the
 * PDFReusable interface due to limitations in the document handling of the
 * PDFBox when the signed pdf is saved.
 * 
 * @author MA_KN
 * 
 */
public class PDFSigner implements SignatureInterface {
	public static final String KEY_STORE_TYPE = "JKS";

	protected static final BouncyCastleProvider PROVIDER = new BouncyCastleProvider();

	protected PrivateKey privKey;
	protected List<Certificate> certificateChain;

	private static final String JCA_SIGN_ALGORITHM = "SHA256withRSA";

	public PDFSigner() throws Exception {
		KeyStore keystore = KeyStore.getInstance(KEY_STORE_TYPE);
		keystore.load(new FileInputStream("resources/keystore_cust.jks"),
				"custpass".toCharArray());
		privKey = (PrivateKey) keystore.getKey("customer",
				"custpass".toCharArray());
		certificateChain = Arrays.asList(keystore
				.getCertificateChain("customer"));
	}

	public void sign(File document, File outputDocument, int page)
			throws Exception {

	    FileInputStream fis = new FileInputStream(document);
	    FileOutputStream fos = new FileOutputStream(outputDocument);
	    
	    byte[] buffer = new byte[8 * 1024];
	    int c;
	    while ((c = fis.read(buffer)) != -1)
	    {
	      fos.write(buffer, 0, c);
	    }
	    fis.close();
	    fis = new FileInputStream(outputDocument);
	    
		// load document
		PDDocument pdfDoc = null;
		try {
			pdfDoc = PDDocument.load(new FileInputStream(document));

			// create signature dictionary
			PDSignature signatureField = new PDSignature();
			signatureField.setFilter(PDSignature.FILTER_ADOBE_PPKLITE); // default
			// Adobe
			// filter
			// subfilter for basic and PAdES Part 2 signatures
			signatureField
					.setSubFilter(PDSignature.SUBFILTER_ADBE_PKCS7_DETACHED);

			// the signing date, needed for a valid signature
			signatureField.setSignDate(Calendar.getInstance());

			// register signature dictionary and sign interface
			signatureField.setLocation(""); // usage?
			signatureField.setName("");
			signatureField.setReason("");

			SignatureOptions signOptions = new SignatureOptions();
			signOptions.setPage(page);
			signOptions.setVisualSignature(new FileInputStream(
					"resources/visual_signature_templ.pdf"));

			pdfDoc.addSignature(signatureField, this, signOptions);

			// TODO: the incremental write to the file system is
			// currently necessary due to a bug -> should be solved in PDFBOX
			// 1.7.0 -> https://issues.apache.org/jira/browse/PDFBOX-1125
			pdfDoc.saveIncremental(fis, fos);
		} finally {
			if (pdfDoc != null) {
				pdfDoc.close();
			}
		}
	}

	/**
	 * Do not use this method. It is a callback method that is necessary for the
	 * PDFBox save method.
	 */
	@Override
	public byte[] sign(InputStream content) throws SignatureException,
			IOException {
		CMSProcessableInputStream input = new CMSProcessableInputStream(content);
		CMSSignedDataGenerator gen = new CMSSignedDataGenerator();

		try {
			ContentSigner signer = new JcaContentSignerBuilder(
					JCA_SIGN_ALGORITHM).setProvider(PROVIDER).build(privKey);

			gen.addSignerInfoGenerator(new JcaSignerInfoGeneratorBuilder(
					new JcaDigestCalculatorProviderBuilder().setProvider(
							PROVIDER).build()).build(signer,
					(X509Certificate) certificateChain.get(0)));

			gen.addCertificates(new JcaCertStore(certificateChain));
			CMSSignedData signedData = gen.generate(input, PROVIDER);
			return signedData.getEncoded();

		} catch (Exception e) {
			throw new SignatureException(e);
		}
	}

}

/**
 * Necessary because the bouncycastle <code>CMSProcessableInputStream</code>
 * implementation is private to the package.
 */
class CMSProcessableInputStream implements CMSProcessable {

	InputStream in;

	public CMSProcessableInputStream(InputStream is) {
		in = is;
	}

	@Override
	public Object getContent() {
		return in;
	}

	@Override
	public void write(OutputStream out) throws IOException, CMSException {
		// read the content only one time
		byte[] buffer = new byte[8192];
		int read;
		while ((read = in.read(buffer)) != -1) {
			out.write(buffer, 0, read);
		}
		in.close();
	}
}